#!/usr/bin/env python
#
# test_webpage.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import six

# python 3
if six.PY3:
    import unittest.mock as mock
# python 2
elif six.PY2:
    import mock

import fsleyes_widgets.utils.webpage as webpage

def test_fileToUrl():

    # fname, expected
    testcases = [
        ('/home/blah/file.html', 'file:///home/blah/file.html'),
        ('file.html',            'file:///file.html'),
        ('./file.html',          'file:///file.html'),
        ('../file.html',         'file:///../file.html'),
    ]

    # urljoin has changed in python3 to make paths absolute
    if six.PY3:
        testcases[-1] = ('../file.html', 'file:///file.html')

    for fname, expected in testcases:
        assert webpage.fileToUrl(fname) == expected


def test_openPage():

    opened = [None]

    def mockOpen(page):
        opened[0] = page

    tomock = 'fsleyes_widgets.utils.webpage.webbrowser.open'
    toopen = 'fakeUrl'
    with mock.patch(tomock, side_effect=mockOpen):

        webpage.openPage(toopen)

    assert opened[0] == toopen


def test_openFile():

    opened = [None]

    def mockOpen(page):
        opened[0] = page

    tomock   = 'fsleyes_widgets.utils.webpage.webbrowser.open'
    toopen   = '/path/to/file.html'
    expected = webpage.fileToUrl(toopen)

    with mock.patch(tomock, side_effect=mockOpen):
        webpage.openFile(toopen)

    assert opened[0] == expected



#!/usr/bin/env python
#
# __init__.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

from __future__ import print_function

import time

import numpy as np

import wx

def compare_images(img1, img2, threshold):
    """Compares two images using the euclidean distance in RGB space
    between pixels. Returns a tuple containing:

     - A boolean value indicating whether the test passed (the images
       were the same).

     - The sum of the normalised RGB distance between all pixels.
    """

    # Discard alpha values
    img1 = img1[:, :, :3]
    img2 = img2[:, :, :3]

    if img1.shape != img2.shape:
        return False, 0


    flat1   = img1.reshape(-1, 3)
    flat2   = img2.reshape(-1, 3)

    dist    = np.sqrt(np.sum((flat1 - flat2) ** 2, axis=1))
    dist    = dist.reshape(img1.shape[:2])
    dist    = dist / np.sqrt(3 * 255 * 255)

    ttlDiff = np.sum(dist)

    passed = ttlDiff <= threshold

    return passed, ttlDiff


def run_with_wx(func, *args, **kwargs):

    propagateRaise = kwargs.pop('propagateRaise', True)
    startingDelay  = kwargs.pop('startingDelay',  500)
    finishingDelay = kwargs.pop('finishingDelay', 500)
    callAfterApp   = kwargs.pop('callAfterApp',   None)

    result = [None]
    raised = [None]

    app    = wx.App()
    frame  = wx.Frame(None)

    if callAfterApp is not None:
        callAfterApp()

    def wrap():

        try:
            if func is not None:
                result[0] = func(*args, **kwargs)

        except Exception as e:
            print(e)
            raised[0] = e

        finally:
            def finish():
                frame.Destroy()
                app.ExitMainLoop()
            wx.CallLater(finishingDelay, finish)

    frame.Show()

    wx.CallLater(startingDelay, wrap)

    app.MainLoop()

    if raised[0] and propagateRaise:
        raise raised[0]

    return result[0]


# Under GTK, a single call to
# yield just doesn't cut it
def realYield():
    for i in range(10):
        wx.Yield()
        time.sleep(0.01)


# stype:
#   0 for single click
#   1 for double click
#   2 for separatemouse down/up events
def simclick(sim, target, btn=wx.MOUSE_BTN_LEFT, pos=None, stype=0):

    w, h = target.GetClientSize().Get()
    x, y = target.GetScreenPosition()

    if pos is None:
        pos = [0.5, 0.5]

    x += w * pos[0]
    y += h * pos[1]

    sim.MouseMove(x, y)
    wx.Yield()
    if   stype == 0: sim.MouseClick(btn)
    elif stype == 1: sim.MouseDblClick(btn)
    else:
        sim.MouseDown(btn)
        sim.MouseUp(btn)
    realYield()


def simtext(sim, target, text, enter=True):
    target.SetFocus()
    target.SetValue(text)
    if enter: sim.KeyDown(wx.WXK_RETURN)
    realYield()


def simkey(sim, target, key, down=True, up=False):
    if target is not None: target.SetFocus()
    if down:               sim.KeyDown(key)
    if up:                 sim.KeyUp(key)
    realYield()



#!/usr/bin/env python
#
# autotextctrl.py - The AutoTextCtrl class.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`AutoTextCtrl` class, an alternative to the
``wx.TextCtrl``, which has auto-completion capability.

I wrote this class because ``wx.TextCtrl`` auto-completion does not work under
OSX, and the ``wx.ComboBox`` does not give me enough fine-grained control with
respect to managing focus.
"""


import logging

import wx
import wx.lib.newevent as wxevent


log = logging.getLogger(__name__)


class AutoTextCtrl(wx.Panel):
    """The ``AutoTextCtrl`` class is essentially a ``wx.TextCtrl`` which is able
    to dynamically show a list of options to the user, with a
    :class:`AutoCompletePopup`.
    """


    def __init__(self, parent, style=0):
        """Create an ``AutoTextCtrl``.

        :arg parent: The ``wx`` parent object.
        :arg style:  Can be :data:`ATC_CASE_SENSITIVE` to restrict the
                     auto-completion options to case sensitive matches.
        """

        self.__caseSensitive = style & ATC_CASE_SENSITIVE

        wx.Panel.__init__(self, parent)

        self.__textCtrl = wx.TextCtrl(self, style=wx.TE_PROCESS_ENTER)
        self.__sizer    = wx.BoxSizer(wx.HORIZONTAL)

        self.__sizer.Add(self.__textCtrl, flag=wx.EXPAND, proportion=1)
        self.SetSizer(self.__sizer)

        # The takeFocus flag is set by SetTakeFocus,
        # and used in __showPopup. The options array
        # contains the auto complete options.
        self.__takeFocus = False
        self.__options   = []

        self.__textCtrl.Bind(wx.EVT_TEXT,        self.__onText)
        self.__textCtrl.Bind(wx.EVT_LEFT_DCLICK, self.__onDoubleClick)
        self.__textCtrl.Bind(wx.EVT_TEXT_ENTER,  self.__onEnter)
        self.__textCtrl.Bind(wx.EVT_KEY_DOWN,    self.__onKeyDown)
        self.__textCtrl.Bind(wx.EVT_SET_FOCUS,   self.__onSetFocus)
        self           .Bind(wx.EVT_SET_FOCUS,   self.__onSetFocus)


    def __onSetFocus(self, ev):
        """Called when this ``AutoTextCtrl`` or any of its children gains
        focus. Makes sure that the text control insertion point is at the
        end of its current contents.
        """

        ev.Skip()

        log.debug('Text control gained focus: {}'.format(
            wx.Window.FindFocus()))

        # Under wx/GTK, when a text control gains focus,
        # it seems to select its entire contents, meaning
        # that when the user types  something, the current
        # contents are replaced with the new contents. To
        # prevent this, here we make sure that no text is
        # selected, and the insertion point is at the end
        # of the current contents.
        text = self.__textCtrl.GetValue()
        self.__textCtrl.SetSelection(len(text) - 1, len(text) - 1)
        self.__textCtrl.SetInsertionPointEnd()


    @property
    def textCtrl(self):
        """Returns a reference to the internal ``wx.TextCtrl``. """
        return self.__textCtrl


    def AutoComplete(self, options):
        """Set the list of options to be shown to the user. """
        self.__options = list(options)


    def GetValue(self):
        """Returns the current value shown on this ``AutoTextCtrl``. """
        return self.__textCtrl.GetValue()


    def SetValue(self, value):
        """Sets the current value shown on this ``AutoTextCtrl``.

        .. note:: Calling this method will result in an ``wx.EVT_TEXT``
                  event being generated - use :meth:`ChangeValue` if you
                  do not want this to occur.
        """
        self.__textCtrl.SetValue(value)


    def ChangeValue(self, value):
        """Sets the current value shown on this ``AutoTextCtrl``. """
        self.__textCtrl.ChangeValue(value)


    def GetInsertionPoint(self):
        """Returns the cursor location in this ``AutoTextCtrl``. """
        return self.__textCtrl.GetInsertionPoint()


    def SetInsertionPoint(self, idx):
        """Sets the cursor location in this ``AutoTextCtrl``. """
        self.__textCtrl.SetInsertionPoint(idx)


    def GenEnterEvent(self):
        """Programmatically generates an :data:`EVT_ATC_TEXT_ENTER` event. """
        self.__onEnter(None)


    def SetTakeFocus(self, takeFocus):
        """If ``takeFocus`` is ``True``, this ``AutoTextCtrl`` will give
        itself focus when its ``AutoCompletePopup`` is closed.
        """
        self.__takeFocus = takeFocus


    def __onKeyDown(self, ev):
        """Called on ``EVT_KEY_DOWN`` events in the text control. """

        enter = wx.WXK_RETURN
        key   = ev.GetKeyCode()

        log.debug('Key event on text control: {}'.format(key))

        # Make sure the event is propagated
        # up the window hierarchy, if we skip it
        ev.ResumePropagation(wx.EVENT_PROPAGATE_MAX)

        if key != enter:
            ev.Skip()
            return

        if self.GetValue() == '':
            log.debug('Enter/right arrow - displaying all options')
            self.__showPopup('')

        # Let the text control handle the event normally
        else:
            ev.Skip()


    def __onDoubleClick(self, ev):
        """Called when the user double clicks in this ``AutoTextCtrl``.
        Creates an :class:`AutoCompletePopup`.
        """
        log.debug('Double click on text control - simulating text entry')
        self.__onText(None)


    def __onText(self, ev):
        """Called when the user changes the text shown on this ``AutoTextCtrl``.
        Creates an :class:`AutoCompletePopup`.
        """

        text = self.__textCtrl.GetValue()
        log.debug('Text - displaying options matching "{}"'.format(text))
        self.__showPopup(text)


    def __onEnter(self, ev):
        """Called when the user presses enter in this ``AutoTextCtrl``. Generates
        an :data:`EVT_ATC_TEXT_ENTER` event.
        """
        value = self.__textCtrl.GetValue()
        ev = AutoTextCtrlEnterEvent(text=value)

        log.debug('Enter - generating ATC enter '
                  'event (text: "{}")'.format(value))

        wx.PostEvent(self, ev)


    def __showPopup(self, text):
        """Creates an :class:`AutoCompletePopup` which displays a list of
        auto-completion options, matching the given prefix text, to the user.

        The popup is not displayed if there are no options with the given
        prefix.
        """

        text            = text.strip()
        style           = 0

        if self.__caseSensitive:
            style |= ATC_CASE_SENSITIVE

        popup = AutoCompletePopup(
            self,
            self,
            text,
            self.__options,
            style)

        if popup.GetCount() == 0:
            popup.Destroy()
            return

        # Don't take focus unless the AutoCompletePopup
        # tells us to (it will call the SetTakeFocus method)
        self.__takeFocus = False

        # Make sure we get the focus back
        # when the popup is destroyed
        def refocus(ev):

            # A call to Raise is required under
            # GTK, as otherwise the main window
            # won't be given focus.
            if wx.Platform == '__WXGTK__':
                self.GetTopLevelParent().Raise()

            if self.__takeFocus:
                self.__textCtrl.SetFocus()

        popup.Bind(wx.EVT_WINDOW_DESTROY, refocus)

        # The popup has its own textctrl - we
        # position the popup so that its textctrl
        # is displayed on top of our textctrl,
        # with the option list underneath.
        posx, posy = self.__textCtrl.GetScreenPosition().Get()

        popup.SetSize((-1, -1))
        popup.SetPosition((posx,  posy))
        popup.Show()


ATC_CASE_SENSITIVE = 1
"""Syle flag for use with the :class:`AutoTextCtrl` class. If set, the
auto-completion pattern matching will be case sensitive.
"""


_AutoTextCtrlEnterEvent, _EVT_ATC_TEXT_ENTER = wxevent.NewEvent()


EVT_ATC_TEXT_ENTER = _EVT_ATC_TEXT_ENTER
"""Identifier for the :data:`AutoTextCtrlEnterEvent`, which is generated
when the user presses enter in an :class:`AutoTextCtrl`.
"""


AutoTextCtrlEnterEvent = _AutoTextCtrlEnterEvent
"""Event generated when the user presses enter in an :class:`AutoTextCtrl`.
Contains a single attribute, ``text``, which contains the text in the
``AutoTextCtrl``.
"""


class AutoCompletePopup(wx.Frame):
    """The ``AutoCompletePopup`` class is used by the :class:`AutoTextCtrl`
    to display a list of completion options to the user.
    """

    def __init__(self, parent, atc, text, options, style=0):
        """Create an ``AutoCompletePopup``.

        :arg parent:  The ``wx`` parent object.
        :arg atc:     The :class:`AutoTextCtrl` that is using this popup.
        :arg text:    Initial text value.
        :arg options: A list of all possible auto-completion options.
        :arg style:   Set to :data:`ATC_CASE_SENSITIVE` to make the
                      pattern matching case sensitive.
        """

        wx.Frame.__init__(self, parent, style=wx.NO_BORDER)

        self.__caseSensitive = style & ATC_CASE_SENSITIVE
        self.__atc           = atc
        self.__options       = options
        self.__textCtrl      = wx.TextCtrl(self,
                                           value=text,
                                           style=wx.TE_PROCESS_ENTER)
        self.__listBox       = wx.ListBox( self,
                                           style=(wx.LB_SINGLE))

        self.__listBox.Set(self.__getMatches(text))

        self.__sizer = wx.BoxSizer(wx.VERTICAL)
        self.__sizer.Add(self.__textCtrl, flag=wx.EXPAND)
        self.__sizer.Add(self.__listBox,  flag=wx.EXPAND, proportion=1)
        self.SetSizer(self.__sizer)

        self.Layout()
        self.Fit()

        self.__textCtrl.Bind(wx.EVT_TEXT,           self.__onText)
        self.__textCtrl.Bind(wx.EVT_TEXT_ENTER,     self.__onEnter)
        self.__textCtrl.Bind(wx.EVT_KEY_DOWN,       self.__onKeyDown)
        self.__listBox .Bind(wx.EVT_KEY_DOWN,       self.__onListKeyDown)
        self.__listBox .Bind(wx.EVT_LISTBOX_DCLICK, self.__onListMouseDblClick)

        # Under GTK, the SetFocus/KillFocus event
        # objects often don't have a reference to
        # the window that received/is about to
        # receive focus. In particular, if the
        # list box is clicked, a killFocus event
        # is triggered, but the list box is not
        # passed in. So on mouse down events, we
        # force the list box to have focus.
        if wx.Platform == '__WXGTK__':
            self.__listBox .Bind(wx.EVT_LEFT_DOWN,  self.__onListMouseDown)
            self.__listBox .Bind(wx.EVT_RIGHT_DOWN, self.__onListMouseDown)

        self           .Bind(wx.EVT_KILL_FOCUS,     self.__onKillFocus)
        self.__textCtrl.Bind(wx.EVT_KILL_FOCUS,     self.__onKillFocus)
        self.__listBox .Bind(wx.EVT_KILL_FOCUS,     self.__onKillFocus)

        self           .Bind(wx.EVT_SET_FOCUS,     self.__onSetFocus)
        self.__textCtrl.Bind(wx.EVT_SET_FOCUS,     self.__onSetFocus)
        self.__listBox .Bind(wx.EVT_SET_FOCUS,     self.__onSetFocus)


    def GetCount(self):
        """Returns the number of auto-completion options currently available.
        """
        return self.__listBox.GetCount()


    def __onSetFocus(self, ev):
        """Called when this ``AutoCompletePopup`` or any of its children gains
        focus. Makes sure that the text control insertion point is at the end
        of its current contents.
        """

        ev.Skip()

        log.debug('Popup gained focus: {}'.format(ev.GetWindow()))

        # See note in AutoTextCtrl.__onSetFocus
        text = self.__textCtrl.GetValue()
        self.__textCtrl.SetSelection(len(text) - 1, len(text) - 1)
        self.__textCtrl.SetInsertionPointEnd()


    def __onKillFocus(self, ev):
        """Called when this ``AutoCompletePopup`` loses focus. Calls
        :meth:`__destroy`.
        """

        ev.Skip()

        focused = ev.GetWindow()

        log.debug('Kill focus event on popup: {}'.format(focused))

        objs = (self, self.__textCtrl, self.__listBox)

        if focused not in objs:
            log.debug('Focus lost - destroying popup')
            self.__destroy(False, False)


    def __destroy(self, genEnter=True, returnFocus=True):
        """Called by various event handlers. Copies the current value in
        this ``AutoCompletePopup`` to the owning :class:`AutoTextCtrl`,
        and then (asynchronously) destroys this ``AutoCompletePopup``.
        """
        value = self.__textCtrl.GetValue()
        idx   = self.__textCtrl.GetInsertionPoint()
        atc   = self.__atc

        # Under wx/GTK, we might still receive focus
        # events, which will trigger another call to
        # __destroy. So we remove all callbacks to
        # prevent this from happening.
        self           .Bind(wx.EVT_SET_FOCUS,  None)
        self.__textCtrl.Bind(wx.EVT_SET_FOCUS,  None)
        self.__listBox .Bind(wx.EVT_SET_FOCUS,  None)
        self           .Bind(wx.EVT_KILL_FOCUS, None)
        self.__textCtrl.Bind(wx.EVT_KILL_FOCUS, None)
        self.__listBox .Bind(wx.EVT_KILL_FOCUS, None)

        atc.ChangeValue(      value)
        atc.SetInsertionPoint(idx)

        # Tell the atc whether or not it
        # should take the focus when this
        # popup is destroyed.
        atc.SetTakeFocus(returnFocus)

        if genEnter:
            atc.GenEnterEvent()

        def destroy():
            try:
                self.Close()
                self.Destroy()

            except wx.PyDeadObjectError:
                pass

        wx.CallAfter(destroy)


    def __getMatches(self, prefix):
        """Returns a list of auto-completion options which match the given
        prefix.
        """

        prefix  = prefix.strip()
        options = self.__options

        if not self.__caseSensitive:
            prefix  = prefix.lower()
            options = [o.lower() for o in options]

        matches = [o.startswith(prefix) for o in options]

        return [o for o, m in zip(self.__options, matches) if m]


    def __onKeyDown(self, ev):
        """Called on an ``EVT_KEY_DOWN`` event from the text control. """

        up    = wx.WXK_UP
        down  = wx.WXK_DOWN
        esc   = wx.WXK_ESCAPE
        enter = wx.WXK_RETURN
        key   = ev.GetKeyCode()

        log.debug('Key down event on popup text control: {}'.format(key))

        if key not in (up, down, enter, esc):
            ev.ResumePropagation(wx.EVENT_PROPAGATE_MAX)
            ev.Skip()
            return

        # Absorb the up arrow
        if key == up:
            return

        # The user hitting enter/escape will result
        # in this popup being destroyed
        if key in (esc, enter):
            log.debug('Enter/escape on popup text '
                      'control - destroying popup')
            self.__destroy(key == enter)
            return

        # If the user hits the down
        # arrow, focus the listbox
        self.__listBox.SetFocus()
        self.__listBox.SetSelection(0)


    def __onText(self, ev):
        """Called on an ``EVT_TEXT`` event from the text control."""

        text    = self.__textCtrl.GetValue().strip()
        matches = self.__getMatches(text)

        if text == '' or len(matches) == 0:
            log.debug('Text on popup text control ("{}") - '
                      'no matches, destroying popup'.format(text))
            self.__destroy(False)
        else:
            log.debug('Text on popup text control ("{}") - '
                      'displaying {} matches'.format(text, len(matches)))
            self.__listBox.Set(matches)


    def __onEnter(self, ev):
        """Called on an ``EVT_TEXT_ENTER`` event from the text control."""

        log.debug('Enter on popup text control - destroying popup')
        self.__destroy()


    def __onListKeyDown(self, ev):
        """Called on an ``EVT_KEY_DOWN`` event from the list box.
        """
        key       = ev.GetKeyCode()
        enter     = wx.WXK_RETURN
        esc       = wx.WXK_ESCAPE
        backspace = wx.WXK_BACK
        delete    = wx.WXK_DELETE
        up        = wx.WXK_UP

        log.debug('Key event on popup list box: {}'.format(key))

        if key not in (enter, esc, up, backspace, delete):
            ev.Skip()
            return

        sel = self.__listBox.GetSelection()
        val = self.__listBox.GetString(sel)

        # If the user pushed the up arrow,
        # and we're at the top of the list,
        # give the focus to the text control
        if key == up:
            if sel == 0:
                log.debug('Up arrow on popup list box - '
                          'shifting focus to text control')
                self.__textCtrl.SetFocus()
                self.__textCtrl.SetInsertionPointEnd()
            else:
                ev.Skip()
            return

        # If the user pushed enter, copy
        # the current list selection to
        # the text control.
        if key == enter:
            log.debug('Enter on popup list box ("{}") - destroying '
                      'popup (and submitting value)'.format(val))
            self.__textCtrl.ChangeValue(val)
            self.__textCtrl.SetInsertionPointEnd()
            genEnter = True

        elif key in (esc, backspace, delete):
            log.debug('Escape on popup list box ("{}") '
                      '- destroying popup'.format(val))
            genEnter = False

        # The user hitting enter or escape
        # will result in this popup being
        # destroyed
        self.__destroy(genEnter)


    def __onListMouseDown(self, ev):
        """Called on GTK when the user clicks in the list box. Forces
        the list box to have focus.
        """
        ev.Skip()
        self.__listBox.SetFocus()


    def __onListMouseDblClick(self, ev):
        """Called when the user double clicks an item in the list box. """
        ev.Skip()

        sel = self.__listBox.GetSelection()
        val = self.__listBox.GetString(sel)

        log.debug('Double click on popup list box ("{}") - '
                  'destroying popup (and submitting value)'.format(val))

        self.__textCtrl.ChangeValue(val)
        self.__textCtrl.SetInsertionPointEnd()
        self.__destroy()



#!/usr/bin/env python
#
# tagtext.py - The StaticTextTag and TextTagPanel classes.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides two classes:

.. autosummary::
   :nosignatures:

   StaticTextTag
   TextTagPanel


.. image:: images/texttagpanel.png
   :scale: 50%
   :align: center
"""


import logging
import random

import wx
import wx.lib.newevent as wxevent

from . import autotextctrl as atc


log = logging.getLogger(__name__)


class StaticTextTag(wx.Panel):
    """The ``StaticTextTag`` class is a ``wx.Panel`` which contains a
    ``StaticText`` control, and a *close* button. The displayed text
    and background colour are configurable.

    When the close button is pushed, an :data:`EVT_STT_CLOSE` is
    generated.
    """


    def __init__(self,
                 parent,
                 text=None,
                 bgColour='#aaaaaa',
                 borderColour='#ffcdcd'):
        """Create a ``StaticTextTag``.

        :arg parent:       The :mod:`wx` parent object.
        :arg text:         Initial text to display.
        :arg bgColour:     Initial background colour.
        :arg borderColour: Initial border colour.
        """

        self.__bgColour     = None
        self.__borderColour = None

        wx.Panel.__init__(self, parent)

        self.__sizer    = wx.BoxSizer(wx.HORIZONTAL)
        self.__closeBtn = wx.StaticText(self,
                                        label='X',
                                        style=(wx.SUNKEN_BORDER         |
                                               wx.ALIGN_CENTRE_VERTICAL |
                                               wx.ALIGN_CENTRE_HORIZONTAL))
        self.__text     = wx.StaticText(self,
                                        style=(wx.ALIGN_CENTRE_VERTICAL |
                                               wx.ALIGN_CENTRE_HORIZONTAL))

        self.__closeBtn.SetFont(self.__closeBtn.GetFont().Smaller())

        self.__sizer.Add(self.__closeBtn,
                         border=2,
                         flag=(wx.EXPAND |
                               wx.LEFT   |
                               wx.TOP    |
                               wx.BOTTOM))
        self.__sizer.Add(self.__text,
                         border=2,
                         flag=(wx.EXPAND |
                               wx.RIGHT  |
                               wx.TOP    |
                               wx.BOTTOM))
        self.SetSizer(self.__sizer)

        self           .SetBackgroundColour(bgColour)
        self           .SetBorderColour(    borderColour)
        self.__closeBtn.SetForegroundColour('#404040')
        self.SetText(text)

        self.__closeBtn.Bind(wx.EVT_LEFT_UP,     self.__onCloseButton)
        self.__closeBtn.Bind(wx.EVT_SET_FOCUS,   self.__onSetFocus)
        self.__text    .Bind(wx.EVT_SET_FOCUS,   self.__onSetFocus)
        self           .Bind(wx.EVT_SET_FOCUS,   self.__onSetFocus)
        self.__closeBtn.Bind(wx.EVT_KILL_FOCUS,  self.__onKillFocus)
        self.__text    .Bind(wx.EVT_KILL_FOCUS,  self.__onKillFocus)
        self           .Bind(wx.EVT_KILL_FOCUS,  self.__onKillFocus)


    def __str__(self):
        """Returns a string representation of this ``StaticTextTag``. """
        return 'StaticTextTag(\'{}\')'.format(self.GetText())


    @property
    def closeButton(self):
        """Returns a reference to the ``StaticText`` control used as the
        close button.
        """
        return self.__closeBtn


    @property
    def text(self):
        """Returns a reference to the ``StaticText`` control used for
        displaying the tag text.
        """
        return self.__text


    def SetBackgroundColour(self, colour):
        """Sets the background colour of this ``StaticTextTag``. """

        wx.Panel.SetBackgroundColour(self,  colour)
        self.__text    .SetBackgroundColour(colour)
        self.__closeBtn.SetBackgroundColour(colour)

        self.__bgColour = colour

        self.Refresh()


    def SetBorderColour(self, colour):
        """Sets the border colour of this ``StaticTextTag``, for when it
        has focus.
        """
        self.__borderColour = colour


    def SetText(self, text):
        """Sets the text shown on this ``StaticTextTag``. """
        if text is None:
            text = ''
        self.__text.SetLabel(text)
        self.Layout()
        self.Fit()


    def GetText(self):
        """Returns the text shown on this ``StaticTextTag``. """
        return self.__text.GetLabel()


    def __onSetFocus(self, ev):
        """Called when this ``StaticTextTag`` gains focus. Changes the border
        colour.
        """
        ev.Skip()
        wx.Panel.SetBackgroundColour(self, self.__borderColour)
        self.Refresh()


    def __onKillFocus(self, ev):
        """Called when this ``StaticTextTag`` loses focus. Clears the border
        colour.
        """
        ev.Skip()
        if ev.GetWindow() not in (self, self.__text, self.__closeBtn):
            wx.Panel.SetBackgroundColour(self, self.__bgColour)
            self.Refresh()


    def __onCloseButton(self, ev):
        """Called when the close button is pushed. Generates an
        :data:`EVT_STT_CLOSE`.
        """

        log.debug('{} close button pressed'.format(str(self)))

        ev = StaticTextTagCloseEvent()
        ev.SetEventObject(self)
        wx.PostEvent(self, ev)


_StaticTextTagCloseEvent, _EVT_STT_CLOSE = wxevent.NewEvent()


EVT_STT_CLOSE = _EVT_STT_CLOSE
"""Identifier for the event generated by a :class:`StaticTextTag` when its
close button is pushed.
"""


StaticTextTagCloseEvent = _StaticTextTagCloseEvent
"""Event object created for an :data:`EVT_STT_CLOSE`. """


class TextTagPanel(wx.Panel):
    """The ``TextTagPanel`` is a panel which contains a control allowing
    the user to add new tags, and a collection of :class:`StaticTextTag`
    controls.

    The ``TextTagPanel`` supports the following styles:

    .. autosummary::

       TTP_ALLOW_NEW_TAGS
       TTP_ADD_NEW_TAGS
       TTP_NO_DUPLICATES
       TTP_CASE_SENSITIVE
       TTP_KEYBOARD_NAV


    The ``TextTagPanel`` generates the following events:

    .. autosummary::

       EVT_TTP_TAG_REMOVED
       EVT_TTP_TAG_ADDED
       EVT_TTP_TAG_SELECT
    """


    def __init__(self, parent, style=None):
        """Create a ``TextTagPanel``.

        :arg parent: The :mod:`wx` parent object.
        :arg style:  Style flags. Defaults to
                     ``TTP_ALLOW_NEW_TAGS | TTP_ADD_NEW_TAGS``.
        """

        wx.Panel.__init__(self, parent)

        if style is None:
            style = TTP_ALLOW_NEW_TAGS | TTP_ADD_NEW_TAGS

        self.__allowNewTags  = style & TTP_ALLOW_NEW_TAGS
        self.__addNewTags    = style & TTP_ADD_NEW_TAGS and self.__allowNewTags
        self.__noDuplicates  = style & TTP_NO_DUPLICATES
        self.__keyboardNav   = style & TTP_KEYBOARD_NAV
        self.__caseSensitive = style & TTP_CASE_SENSITIVE

        self.__allTags     = []
        self.__tagDisplays = {}
        self.__activeTags  = {}
        self.__tagColours  = {}
        self.__tagWidgets  = []

        if self.__caseSensitive: atcStyle = atc.ATC_CASE_SENSITIVE
        else:                    atcStyle = 0

        self.__newTagCtrl = atc.AutoTextCtrl(self, style=atcStyle)
        self.__mainSizer  = wx.BoxSizer( wx.HORIZONTAL)
        self.__tagSizer   = wx.WrapSizer(wx.HORIZONTAL, 2)

        # ^^ the WrapSizer style flags don't
        # seem to have made it into wxPython:
        #
        #     EXTEND_LAST_ON_EACH_LINE = 1
        #     REMOVE_LEADING_SPACES    = 2

        self.__mainSizer.Add(self.__newTagCtrl)
        self.__mainSizer.Add(self.__tagSizer, flag=wx.EXPAND, proportion=1)

        self.__newTagCtrl.Bind(atc.EVT_ATC_TEXT_ENTER, self.__onTextCtrl)

        if self.__keyboardNav:
            self.__newTagCtrl.Bind(wx.EVT_KEY_DOWN, self.__onNewTagKeyDown)

        self.SetSizer(self.__mainSizer)
        self.Layout()


    @property
    def newTagCtrl(self):
        """Returns a reference to the :class:`.AutoTextCtrl`. """
        return self.__newTagCtrl


    @property
    def tags(self):
        """Returns a list containing all :class:`StaticTextTag` widgets. """
        return list(self.__tagWidgets)


    def FocusNewTagCtrl(self):
        """Gives focus to the new tag control (an :class:`.AutoTextCtrl`).
        """
        self.__newTagCtrl.SetFocus()


    def SelectTag(self, tag):
        """Gives focus to the :class:`StaticTextTag` control with the
        specified tag, if it exists.
        """
        tagIdx = self.GetTagIndex(tag)
        self.__tagWidgets[tagIdx].SetFocus()


    def SetOptions(self, options, colours=None):
        """Sets the tag options made available to the user via the
        :class:`.AutoTextCtrl`.

        :arg options: A sequence of tags that the user can choose from.
        :arg colours: A sequence of corresponding colours for each tag.
        """

        origOptions = options

        if not self.__caseSensitive:
            lowered = [o.lower() for o in options]
            uniq    = []
            orig    = []

            for i, o in enumerate(lowered):
                if o not in uniq:
                    uniq.append(o)
                    orig.append(origOptions[i])

            options     = uniq
            origOptions = orig

        self.__allTags     = list(options)
        self.__tagDisplays = {o : oo for (o, oo) in zip(options, origOptions)}

        self.__updateNewTagOptions()

        if colours is not None:

            for option, colour in zip(options, colours):
                self.SetTagColour(option, colour)

        # TODO delete any active tags
        #      that are no longer valid?


    def GetOptions(self):
        """Returns a list of all the tags that are currently available to the
        user.
        """
        return [self.__tagDisplays[o] for o in self.__allTags]


    def AddTag(self, tag, colour=None):
        """Add a new :class:`StaticTextTag` to this ``TextTagPanel``.

        :arg tag:    The tag text.
        :arg colour: The tag background colour.
        """

        origTag = tag

        if not self.__caseSensitive:
            tag = tag.lower()

        if colour is None:
            colour = self.__tagColours.get(tag, None)

        if colour is None:
            colour = [random.randint(100, 255),
                      random.randint(100, 255),
                      random.randint(100, 255)]

        stt = StaticTextTag(self, origTag, colour)

        stt.Bind(EVT_STT_CLOSE, self.__onTagClose)

        if self.__keyboardNav:
            stt.Bind(wx.EVT_LEFT_DOWN, self.__onTagLeftDown)
            stt.Bind(wx.EVT_KEY_DOWN,  self.__onTagKeyDown)

        self.__tagSizer.Add(stt, flag=wx.ALL, border=3)
        self.Layout()
        self.GetParent().Layout()

        if self.__addNewTags and tag not in self.__allTags:
            log.debug('Adding new tag to options: {}'.format(tag))
            self.__allTags.append(tag)
            self.__tagDisplays[tag] = origTag

        self.__tagWidgets.append(stt)
        self.__tagColours[tag] = colour
        self.__activeTags[tag] = self.__activeTags.get(tag, 0) + 1
        self.__updateNewTagOptions()


    def RemoveTag(self, tag):
        """Removes the specified tag. """

        if not self.__caseSensitive:
            tag = tag.lower()

        tagIdx = self.GetTagIndex(tag)
        stt    = self.__tagWidgets[tagIdx]

        self.__tagSizer  .Detach(stt)
        self.__tagWidgets.remove(stt)

        count = self.__activeTags[tag]

        if count == 1: self.__activeTags.pop(tag)
        else:          self.__activeTags[tag] = count - 1

        stt.Destroy()
        self.Layout()
        self.GetParent().Layout()

        self.__updateNewTagOptions()


    def GetTags(self):
        """Returns a list containing all active tags in this ``TextTagPanel``.
        """
        return [stt.GetText() for stt in self.__tagWidgets]


    def ClearTags(self):
        """Removes all tags from this ``TextTagPanel``. """
        for tag in list(self.__tagWidgets):
            self.RemoveTag(tag.GetText())


    def GetTagIndex(self, tag):
        """Returns the index of the specified tag. """

        tags = self.GetTags()

        if not self.__caseSensitive:
            tag  = tag.lower()
            tags = [t.lower() for t in tags]

        for i, t in enumerate(tags):
            if t == tag:
                return i

        raise IndexError('Unknown tag: "{}"'.format(tag))


    def TagCount(self):
        """Returns the number of tags currently visible. """
        return len(self.__tagWidgets)


    def HasTag(self, tag):
        """Returns ``True`` if the given tag is currently shown, ``False``
        otherwise.
        """
        if not self.__caseSensitive:
            tag = tag.lower()

        return tag in self.__activeTags


    def GetTagColour(self, tag):
        """Returns the background colour of the specified ``tag``, or ``None``
        if there is no default colour for the tag.
        """
        if not self.__caseSensitive:
            tag = tag.lower()

        return self.__tagColours.get(tag, None)


    def SetTagColour(self, tag, colour):
        """Sets the background colour on all :class:`StaticTextTag` items
        which have the given tag text.
        """

        if not self.__caseSensitive:
            tag = tag.lower()

        if tag not in self.__allTags:
            return

        self.__tagColours[tag] = colour

        for stt in self.__tagWidgets:

            if self.__caseSensitive: sttText = stt.GetText()
            else:                    sttText = stt.GetText().lower()

            if sttText == tag:
                stt.SetBackgroundColour(colour)


    def __selectTag(self, stt):
        """Called by event handlers which listen for mouse/keyboard activity
        on :class:`StaticTextTag` widgets. Focuses the given ``StaticTextTag``,
        and generates an :data:`EVT_TTP_TAG_SELECT` event.
        """

        tag = stt.GetText()
        stt.SetFocus()

        log.debug('Posting tag select event ("{}")'.format(tag))

        ev = TextTagPanelTagSelectEvent(tag=tag)
        ev.SetEventObject(self)
        wx.PostEvent(self, ev)


    def __onTagLeftDown(self, ev):
        """Called on left mouse down events on :class:`StaticTextTag`
        objects (only if the :data:`TTP_KEYBOARD_NAV` style is set).
        Gives the tag focus.
        """

        stt = ev.GetEventObject()
        tag = stt.GetText()

        log.debug('Mouse down on tag: "{}"'.format(tag))
        self.__selectTag(stt)


    def __onTagClose(self, ev):
        """Called when the user pushes the close button on a
        :class:`StaticTextTag`. Removes the tag, and generates a
        :data:`EVT_TTP_TAG_REMOVED` event.
        """
        stt = ev.GetEventObject()
        tag = stt.GetText()

        idx = self.GetTagIndex(tag)

        self.RemoveTag(tag)

        if len(self.__tagWidgets) == 0:
            self.FocusNewTagCtrl()
        else:
            if idx == len(self.__tagWidgets):
                idx -= 1

            self.__tagWidgets[idx].SetFocus()

        log.debug('Tag removed: "{}"'.format(tag))

        ev = TextTagPanelTagRemovedEvent(tag=tag)
        ev.SetEventObject(self)
        wx.PostEvent(self, ev)


    def __onNewTagKeyDown(self, ev):
        """Called on key down events from the new tag control (if the
        :data:`TTP_KEYBOARD_NAV` style is set). If the right
        arrow key is pushed, the first :class:`StaticTextTag` is given input
        focus.
        """

        key = ev.GetKeyCode()

        log.debug('TextTagPanel key down [new tag control] ({})'.format(key))

        # Only process right arrows if the text
        # control cursor is on the far right
        value  = self.__newTagCtrl.GetValue()
        cursor = self.__newTagCtrl.GetInsertionPoint()

        if key != wx.WXK_RIGHT         or \
           len(self.__tagWidgets) == 0 or \
           cursor != len(value):
            ev.Skip()
            return

        log.debug('Right arrow key on new tag control - focusing tags')

        self.__selectTag(self.__tagWidgets[0])


    def __onTagKeyDown(self, ev):
        """Called on key down events from a :class:`StaticTextTag` object. If
        the left/right arrow keys are pushed, the focus is shifted accordingly.
        """

        left      = wx.WXK_LEFT
        right     = wx.WXK_RIGHT
        delete    = wx.WXK_DELETE
        backspace = wx.WXK_BACK
        key       = ev.GetKeyCode()
        stt       = ev.GetEventObject()

        log.debug('TextTagPanel key event ({})'.format(key))

        if key not in (left, right, delete, backspace):
            ev.ResumePropagation(wx.EVENT_PROPAGATE_MAX)
            ev.Skip()
            return

        if key in (delete, backspace):
            self.__onTagClose(ev)
            return

        sttIdx = self.__tagWidgets.index(stt)

        if   key == left:  sttIdx -= 1
        elif key == right: sttIdx += 1

        if sttIdx == -1:
            log.debug('Arrow key on tag ({}) - focusing new '
                      'tag control'.format(stt.GetText()))
            self.FocusNewTagCtrl()
            return

        elif sttIdx == len(self.__tagWidgets):
            ev.Skip()
            return

        log.debug('Arrow key on tag ({}) - selecting '
                  'adjacent tag ({})'.format(
                      stt.GetText(), self.__tagWidgets[sttIdx].GetText()))

        self.__selectTag(self.__tagWidgets[sttIdx])


    def __onTextCtrl(self, ev):
        """Called when the user enters a new value via the ``TextCtrl`` (if
        this ``TextTagPanel`` allows new tags). Adds the new tag, and generates
        an :data:`EVT_TTP_TAG_ADDED` event.
        """

        tag = self.__newTagCtrl.GetValue().strip()

        self.__newTagCtrl.ChangeValue('')

        if tag == '':
            return

        # If we don't care about case, and
        # this is a known option, use the
        # 'display' version of the tag.
        origTag = tag
        if not self.__caseSensitive:
            tag     = tag.lower()
            origTag = self.__tagDisplays.get(tag, origTag)

        if self.__noDuplicates and self.HasTag(tag):
            log.debug('New tag {} ignored (noDuplicates is True)'.format(tag))
            return

        if not self.__allowNewTags and tag not in self.__allTags:
            log.debug('New tag {} ignored (allowNewTags is False)'.format(tag))
            return

        log.debug('New tag from text control: {}'.format(origTag))

        self.__newTagCtrl.Refresh()

        self.AddTag(origTag)

        ev = TextTagPanelTagAddedEvent(tag=origTag)
        ev.SetEventObject(self)
        wx.PostEvent(self, ev)


    def __updateNewTagOptions(self):
        """Updates the options shown on the new tag control."""

        tags = list(self.__allTags)

        if self.__noDuplicates:
            tags = [t for t in tags if t not in self.__activeTags]

        tags = [self.__tagDisplays[t] for t in tags]

        self.__newTagCtrl.AutoComplete(tags)


TTP_ALLOW_NEW_TAGS  = 1
"""Style flag for use with a :class:`TextTagPanel` - if set, the user is able
to type in tag names that are not known by the :class:`.AutoTextCtrl`.
"""


TTP_ADD_NEW_TAGS = 2
"""Style flag for use with a :class:`TextTagPanel` - if set, when the user
types in a tag name that is not known by the ``AutoTextCtrl``, that name is
added to its list of options. This flag only has an effect if the
:data:`TTP_ALLOW_NEW_TAGS` flag is also set.
"""


TTP_NO_DUPLICATES = 4
"""Style flag for use with a :class:`TextTagPanel` - if set, the user will be
prevented from adding the same tag more than once.
"""


TTP_CASE_SENSITIVE = 8
"""Style flag for use with a :class:`TextTagPanel` - if set, the
auto-completion options will be case sensitive. This flag only has an effect
if the :data:`TTP_ALLOW_NEW_TAGS` flag is also set.
"""


TTP_KEYBOARD_NAV = 16
"""Style flag for use with a :class:`TextTagPanel` - if set, the user
can use the left and right arrow keys to move between the new tag control
and the tags and, when a tag is focused can use the delete/backspace keys to
remove it.
"""


_TextTagPanelTagAddedEvent,   _EVT_TTP_TAG_ADDED   = wxevent.NewEvent()
_TextTagPanelTagRemovedEvent, _EVT_TTP_TAG_REMOVED = wxevent.NewEvent()
_TextTagPanelTagSelectEvent,  _EVT_TTP_TAG_SELECT  = wxevent.NewEvent()


EVT_TTP_TAG_ADDED = _EVT_TTP_TAG_ADDED
"""Identifier for the event generated when a tag is added to a
:class:`TextTagPanel`.
"""


TextTagPanelTagAddedEvent = _TextTagPanelTagAddedEvent
"""Event generated when a tag is added to a :class:`TextTagPanel`. A
``TextTagPanelTagAddedEvent`` has a single attribute called ``tag``, which
contains the tag text.
"""


EVT_TTP_TAG_REMOVED = _EVT_TTP_TAG_REMOVED
"""Identifier for the event generated when a tag is removed from a
:class:`TextTagPanel`.
"""


TextTagPanelTagRemovedEvent = _TextTagPanelTagRemovedEvent
"""Event generated when a tag is removed from a :class:`TextTagPanel`. A
``TextTagPanelTagRemovedEvent`` has a single attribute called ``tag``, which
contains the tag text.
"""


EVT_TTP_TAG_SELECT = _EVT_TTP_TAG_SELECT
"""Identifier for the event generated when a tag is selected in a
:class:`TextTagPanel`.
"""


TextTagPanelTagSelectEvent = _TextTagPanelTagSelectEvent
"""Event generated when a tag is selected in  a :class:`TextTagPanel`. A
``TextTagPanelTagSelectEvent`` has a single attribute called ``tag``, which
contains the tag text.
"""



#!/usr/bin/env python
#
# test_runwindow.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import os
import time
import signal
import tempfile

import fsleyes_widgets.utils.runwindow as runwindow

from . import run_with_wx


def _gen_dummy_script():

    script = """#!/bin/bash
    for ((i=0;i<10;i++)); do
        echo "$i"
        echo "err$i" 1>&2
    done
    """

    hd, path = tempfile.mkstemp()

    hd = os.fdopen(hd, 'w')
    hd.write(script)
    hd.close()

    expected   = '\n'.join('{0}\nerr{0}'.format(i) for i in range(10))

    return path, expected


# import logging
# logging.basicConfig()
# logging.getLogger('fsleyes_widgets').setLevel(logging.DEBUG)
def test_ProcessManager_run():

    path, expected = _gen_dummy_script()
    cmd            = ['bash', path]
    result         = [None]
    finishArgs     = [None]

    def runTest():
        import wx

        def onFinish(parent, retcode):
            finishArgs[0] = (parent, retcode)
            result[    0] = rp.text.GetValue()

        frame = wx.GetApp().GetTopWindow()
        rp    = runwindow.RunPanel(frame)
        pm    = runwindow.ProcessManager(cmd, frame, rp, onFinish)

        pm.start()

    run_with_wx(runTest)

    os.remove(path)

    assert expected in result[0]
    assert finishArgs[0][1] == 0



def test_ProcessManager_termProc():

    cmd        = 'sleep 10'.split()
    finishArgs = [None]

    def runTest():
        import wx

        def onFinish(parent, retcode):
            finishArgs[0] = (parent, retcode)

        frame = wx.GetApp().GetTopWindow()
        rp    = runwindow.RunPanel(frame)
        pm    = runwindow.ProcessManager(cmd, frame, rp, onFinish)

        pm.start()

        for i in range(2):
            wx.Yield()
            time.sleep(1)

        pm.termProc()

    run_with_wx(runTest)

    assert finishArgs[0][1] == -signal.SIGTERM


def test_run():

    import wx

    path, expected = _gen_dummy_script()
    cmd            = ['bash', path]
    finishArgs     = [None]

    def runTest():

        frame = wx.GetApp().GetTopWindow()

        def onFinish(parent, retcode):
            finishArgs[0] = (parent, retcode)

        runwindow.run('Tool', cmd, frame, onFinish, modal=False)

    run_with_wx(runTest)

    os.remove(path)

    assert finishArgs[0][1] == 0



#!/usr/bin/env python
#
# test_importall.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


import pkgutil


def test_importall():
    import fsleyes_widgets       as fwidgets
    import fsleyes_widgets.utils as futils

    for _, module, _ in pkgutil.iter_modules(fwidgets.__path__,
                                             'fsleyes_widgets.'):
        __import__(module)

    for _, module, _ in pkgutil.iter_modules(futils.__path__,
                                             'fsleyes_widgets.utils.'):
        __import__(module)



#!/usr/bin/env python
#
# imagepanel.py - A panel for displaying a wx.Image.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module provides the :class:`ImagePanel` class, for displaying a
:class:`wx.Image`.
"""


import wx


class ImagePanel(wx.Panel):
    """A :class:`wx.Panel` which may be used to display a resizeable
    :class:`wx.Image`. The image is scaled to the size of the panel.
    """

    def __init__(self, parent, image=None):
        """Create an ``ImagePanel``.

        If the ``image`` is not passed in here, it can be set later with the
        :meth:`SetImage` method.

        :arg parent: The :mod:`wx` parent object.

        :arg image:  The :class:`wx.Image` object to display.
        """

        wx.Panel.__init__(self, parent)

        self.Bind(wx.EVT_PAINT, self.Draw)
        self.Bind(wx.EVT_SIZE,  self.__onSize)

        self.SetImage(image)


    def SetImage(self, image):
        """Set the image that is displayed on this ``ImagePanel``.

        :arg image: The :class:`wx.Image` object to display.
        """
        self.__image = image


        if image is not None: self.SetMinSize(image.GetSize())
        else:                 self.SetMinSize((0, 0))

        self.Refresh()


    def __onSize(self, ev):
        """Redraw this panel when it is sized, so the image is scaled
        appropriately - see the :meth:`Draw` method.
        """
        self.Refresh()
        ev.Skip()


    def DoGetBestSize(self):
        """Returns the size of the image being displayed.
        """

        if self.__image is None: return (0, 0)
        else:                    return self.__image.GetSize()


    def Draw(self, ev=None):
        """Draws this ``ImagePanel``. The image is scaled to the current panel
        size.
        """

        self.ClearBackground()

        if self.__image is None:
            return

        if ev is None: dc = wx.ClientDC(self)
        else:          dc = wx.PaintDC( self)

        if not dc.IsOk():
            return

        width, height = dc.GetSize().Get()

        if width == 0 or height == 0:
            return

        bitmap = self.__image.Scale(width, height).ConvertToBitmap()

        dc.DrawBitmap(bitmap, 0, 0, False)



#!/usr/bin/env python
#
# test_colourbutton.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import wx

import mock
import pytest

from . import run_with_wx, simclick

import fsleyes_widgets.colourbutton as cb


def test_Create():
    run_with_wx(_test_Create)
def _test_Create():

    frame = wx.GetApp().GetTopWindow()
    btn = cb.ColourButton(frame)

    assert list(btn.GetValue()) == [0, 0, 0, 255]
    btn.SetValue((20, 30, 40))
    assert list(btn.GetValue()) == [20, 30, 40, 255]
    btn.SetValue((20, 30, 40, 50))
    assert list(btn.GetValue()) == [20, 30, 40, 50]


    with pytest.raises(ValueError): btn.SetValue([0])
    with pytest.raises(ValueError): btn.SetValue([0, 1])
    with pytest.raises(ValueError): btn.SetValue([0, 1, 2, 3, 4])
    with pytest.raises(ValueError): btn.SetValue([-1, 0, 2])
    with pytest.raises(ValueError): btn.SetValue([256, 20, 6])



class MockColourDialog(object):

    retval = wx.ID_OK
    colour = [0, 0, 0, 255]

    def __init__(self, parent, colourData):
        self.colourData = colourData

    def ShowModal(self):
        self.colourData.SetColour(MockColourDialog.colour)
        return MockColourDialog.retval

    def GetColourData(self):
        return self.colourData


def test_Event():
    run_with_wx(_test_Event)
def _test_Event():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    btn   = cb.ColourButton(frame)

    colours = [(50, 60, 70),
               (90, 100, 150, 200)]

    result = [None]

    def handler(ev):
        result[0] = ev.colour

    btn.Bind(cb.EVT_COLOUR_BUTTON_EVENT, handler)

    for colour in colours:

        MockColourDialog.colour = colour

        with mock.patch('fsleyes_widgets.colourbutton.wx.ColourDialog',
                        side_effect=MockColourDialog):
            simclick(sim, btn)

        if len(colour) == 3:
            colour = list(colour) + [255]

        assert list(result[0])      == list(colour)
        assert list(btn.GetValue()) == list(colour)

    # Test dialog cancel
    MockColourDialog.retval = wx.ID_CANCEL
    MockColourDialog.colour = (20, 20, 20, 10)

    colour = (150, 160, 170)
    result = [None]
    btn.SetValue(colour)

    with mock.patch('fsleyes_widgets.colourbutton.wx.ColourDialog',
                    side_effect=MockColourDialog):
        simclick(sim, btn)

    assert result[0] is None
    assert list(btn.GetValue()) == list(colour) + [255]



#!/usr/bin/env python
#
# test_autotextctrl.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#


from . import run_with_wx, simclick, realYield

import wx
import fsleyes_widgets.autotextctrl as autott


def sendEvent(target, evType, source=None):

    if source is None:
        source = target

    target.ProcessEvent(wx.CommandEvent(evType, source.GetId()))
    wx.Yield()

# Simple test - programmatically
# set, then retrieve the value
def test_getSet():
    run_with_wx( _test_getSet)
def _test_getSet():

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent)

    atc.ChangeValue('a')
    assert atc.GetValue() == 'a'

    atc.SetValue('b')
    assert atc.GetValue() == 'b'


def test_event():
    run_with_wx( _test_event)
def _test_event():

    called = [None]

    def handler(ev):
        called[0] = ev.text

    sim    = wx.UIActionSimulator()
    parent = wx.GetApp().GetTopWindow()
    atc    = autott.AutoTextCtrl(parent)

    atc.Bind(autott.EVT_ATC_TEXT_ENTER, handler)
    atc.SetFocus()
    realYield()
    sim.Char(ord('a'))
    sim.Char(ord('b'))
    sim.Char(ord('c'))
    sim.KeyDown(wx.WXK_RETURN)
    realYield()

    assert atc.GetValue() == 'abc'
    assert called[0]      == 'abc'


# Make sure that when the cotrol receives
# focus, its insertion point is at the end
def test_onFocus():
    run_with_wx(_test_onFocus)
def _test_onFocus():

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    atc.ChangeValue('a')
    atc.SetInsertionPoint(0)

    sendEvent(atc, wx.wxEVT_KILL_FOCUS)

    assert atc.GetInsertionPoint() == 0

    sendEvent(atc, wx.wxEVT_SET_FOCUS)
    assert atc.GetInsertionPoint() == 1


# Test showing the popup and selecting a value
def test_popup_select1():
    run_with_wx(_test_popup_select1)
def _test_popup_select1():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])
    atc.SetFocus()
    sim.KeyDown(wx.WXK_RETURN)
    realYield()
    sim.KeyDown(wx.WXK_DOWN)
    sim.KeyDown(wx.WXK_DOWN)
    sim.KeyDown(wx.WXK_RETURN)
    realYield()

    assert atc.GetValue() == 'aab'


def test_popup_select2():
    run_with_wx(_test_popup_select2)
def _test_popup_select2():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    atc.SetFocus()
    sim.Char(ord('b'))
    realYield()
    sim.KeyDown(wx.WXK_DOWN)
    sim.KeyDown(wx.WXK_RETURN)
    realYield()

    assert atc.GetValue() == 'bcc'


def test_popup_select3():
    run_with_wx(_test_popup_select3)
def _test_popup_select3():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    atc.SetFocus()
    sim.Char(ord('a'))
    sim.Char(ord('b'))
    realYield()
    sim.KeyDown(wx.WXK_DOWN)
    sim.KeyDown(wx.WXK_RETURN)
    realYield()

    assert atc.GetValue() == 'aba'


def test_popup_cancel():
    run_with_wx(_test_popup_cancel)
def _test_popup_cancel():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    atc.SetFocus()
    sim.KeyDown(wx.WXK_RETURN)
    realYield()
    sim.KeyDown(wx.WXK_DOWN)
    sim.KeyDown(wx.WXK_ESCAPE)
    realYield()

    assert atc.GetValue() == ''


def test_popup_focusback():
    run_with_wx(_test_popup_focusback)
def _test_popup_focusback():

    sim = wx.UIActionSimulator()

    parent = wx.GetApp().GetTopWindow()
    atc = autott.AutoTextCtrl(parent)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    atc.SetFocus()
    sim.KeyDown(wx.WXK_RETURN)
    realYield()
    sim.KeyDown(wx.WXK_DOWN)
    sim.KeyDown(wx.WXK_UP)
    realYield()

    sim.Text('abc')
    sim.KeyDown(wx.WXK_RETURN)
    realYield()

    assert atc.GetValue() == 'abc'


def test_popup_dblclick():
    run_with_wx(_test_popup_dblclick)
def _test_popup_dblclick():
    sim    = wx.UIActionSimulator()
    parent = wx.GetApp().GetTopWindow()
    atc    = autott.AutoTextCtrl(parent)

    atc.AutoComplete(['aaa', 'aab', 'aba', 'bcc'])

    atc.SetFocus()
    sim.KeyDown(wx.WXK_RETURN)
    realYield()

    # Sneakily get a ref to the listbox
    # in the AutoCompletePopup
    realYield()
    listbox = None
    for c in atc.GetChildren():
        if isinstance(c, autott.AutoCompletePopup):

            for pc in c.GetChildren():
                if isinstance(pc, wx.ListBox):
                    listbox = pc
                    break

    simclick(sim, listbox, stype=1, pos=[0.5, 0.05])
    realYield()

    assert atc.GetValue() == 'aaa'



#!/usr/bin/env python
#
# __init__.py - Small utility modules for doing random things.
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""This module contains a collection of small utility modules for doing random
things.
"""



#!/usr/bin/env python
#
# test_rangeslider.py -
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#

import numpy as np

import wx

import mock
import pytest

import fsleyes_widgets.rangeslider as rangeslider

from . import run_with_wx, simclick, simtext, simkey, realYield


def _test_RangePanel_logic(panel):
    panel.SetLimits(100, 1000)
    assert panel.GetLimits() == (100, 1000)
    assert panel.GetMin()    == 100
    assert panel.GetMax()    == 1000
    panel.SetMin(0)
    panel.SetMax(100)
    assert panel.GetLimits() == (0, 100)
    assert panel.GetMin()    == 0
    assert panel.GetMax()    == 100
    panel.SetRange(0, 100)
    panel.SetDistance(5)
    assert panel.GetRange()    == (0, 100)
    assert panel.GetLow()      == 0
    assert panel.GetHigh()     == 100
    assert panel.GetDistance() == 5

    panel.SetLow(5)
    assert np.all(np.isclose(panel.GetRange(), (5, 100)))
    panel.SetHigh(95)
    assert np.all(np.isclose(panel.GetRange(), (5, 95)))
    panel.SetLow(95)
    assert np.all(np.isclose(panel.GetRange(), (92.5, 97.5)))
    panel.SetRange(50, 75)
    panel.SetHigh(51)
    assert np.all(np.isclose(panel.GetRange(), (48, 53)))

    with pytest.raises(ValueError): panel.SetRange( 100, 0)
    with pytest.raises(ValueError): panel.SetLimits(100, 0)
    with pytest.raises(ValueError): panel.SetDistance(101)
    with pytest.raises(ValueError): panel.SetLimits(0, 3)

    # range, expected
    testcases = [
        ((  0, 100), (    0,   100)),
        ((  5, 100), (    5,   100)),
        ((  5,  95), (    5,    95)),
        (( 95,  95), ( 92.5,  97.5)),
        (( 95, 100), (   95,   100)),
        (( 90,  90), ( 87.5,  92.5)),
        ((  0,   0), (    0,     5)),
        ((  0,   5), (    0,     5)),
        ((  0,  10), (    0,    10)),
        ((100, 100), (   95,   100)),
    ]

    for inrange, expected in testcases:
        panel.SetRange(*inrange)
        assert np.all(np.isclose(panel.GetRange(), expected))

    panel.SetRange(0, 100)
    panel.SetLimits(25, 50)
    assert np.all(np.isclose(panel.GetRange(), (25, 50)))

    panel.SetRange(30, 35)
    panel.SetDistance(10)
    assert np.all(np.isclose(panel.GetRange(), (27.5, 37.5)))


def test_RangePanel_logic_slider():
    run_with_wx(_test_RangePanel_logic_slider)
def _test_RangePanel_logic_slider():

    frame = wx.GetApp().GetTopWindow()
    panel = rangeslider.RangePanel(frame, style=rangeslider.RP_SLIDER)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()
    _test_RangePanel_logic(panel)


def test_RangePanel_logic_spin():
    run_with_wx(_test_RangePanel_logic_spin)
def _test_RangePanel_logic_spin():
    frame = wx.GetApp().GetTopWindow()
    panel = rangeslider.RangePanel(frame,
                                   lowLabel='min',
                                   highLabel='max')
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()
    _test_RangePanel_logic(panel)


def test_RangePanel_events_slider():
    run_with_wx(_test_RangePanel_events_slider)
def _test_RangePanel_events_slider():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = rangeslider.RangePanel(frame, style=rangeslider.RP_SLIDER)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()

    handlersCalled = [False] * 3

    def lowHandler(  ev): handlersCalled[0] = True
    def highHandler( ev): handlersCalled[1] = True
    def rangeHandler(ev): handlersCalled[2] = True

    def reset():
        handlersCalled[0] = False
        handlersCalled[1] = False
        handlersCalled[2] = False

    panel.Bind(rangeslider.EVT_LOW_RANGE,  lowHandler)
    panel.Bind(rangeslider.EVT_HIGH_RANGE, highHandler)
    panel.Bind(rangeslider.EVT_RANGE,      rangeHandler)
    panel.SetRange(0, 100)
    panel.SetDistance(5)

    simclick(sim, panel.lowWidget,  pos=[0.25, 0.5])
    assert abs(panel.GetLow() - 25) < 5
    assert  panel.GetHigh() == 100
    assert handlersCalled == [True, False, False]

    reset()
    simclick(sim, panel.highWidget, pos=[0.75, 0.5])
    assert abs(panel.GetLow()  - 25) < 5
    assert abs(panel.GetHigh() - 75) < 5
    assert handlersCalled == [False, True, False]

    reset()
    simclick(sim, panel.highWidget, pos=[0.25, 0.5])
    assert abs(panel.GetLow()  - 20) < 5
    assert abs(panel.GetHigh() - 25) < 5
    assert handlersCalled == [False, False, True]

    reset()
    panel.SetRange(25, 75)
    simclick(sim, panel.lowWidget,  pos=[0.75, 0.5])
    assert abs(panel.GetLow()  - 75) < 5
    assert abs(panel.GetHigh() - 80) < 5
    assert handlersCalled == [False, False, True]

    reset()
    panel.SetRange(25, 75)
    simclick(sim, panel.lowWidget, pos=[0.90, 0.5])
    assert abs(panel.GetLow()  - 90) < 5
    assert abs(panel.GetHigh() - 95) < 5
    assert handlersCalled == [False, False, True]

    reset()
    panel.SetRange(25, 75)
    simclick(sim, panel.highWidget, pos=[0.1, 0.5])
    assert abs(panel.GetLow()  -  5)  < 5
    assert abs(panel.GetHigh() -  10) < 5
    assert handlersCalled == [False, False, True]


def test_RangePanel_events_spin():
    run_with_wx(_test_RangePanel_events_spin)
def _test_RangePanel_events_spin():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = rangeslider.RangePanel(frame)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND)
    frame.SetSizer(sizer)
    frame.Layout()

    handlersCalled = [False] * 3

    def lowHandler(  ev): handlersCalled[0] = True
    def highHandler( ev): handlersCalled[1] = True
    def rangeHandler(ev): handlersCalled[2] = True

    def reset():
        handlersCalled[0] = False
        handlersCalled[1] = False
        handlersCalled[2] = False

    panel.SetRange(0, 100)
    panel.SetDistance(5)

    realYield()

    panel.lowWidget.textCtrl.ChangeValue('25')
    simkey(sim, panel.lowWidget, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (25, 100)))

    panel.highWidget.textCtrl.ChangeValue('75')
    simkey(sim, panel.highWidget, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (25, 75)))

    panel.highWidget.textCtrl.ChangeValue('25')
    simkey(sim, panel.highWidget, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (22.5, 27.5)))

    panel.SetRange(25, 75)
    panel.lowWidget.textCtrl.ChangeValue('75')
    simkey(sim, panel.lowWidget, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (72.5, 77.5)))

    panel.SetRange(25, 75)
    panel.lowWidget.textCtrl.ChangeValue('90')
    simkey(sim, panel.lowWidget, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (90, 95)))

    panel.SetRange(25, 75)
    panel.highWidget.textCtrl.ChangeValue('10')
    simkey(sim, panel.highWidget, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (5, 10)))



def test_RangeSliderSpinPanel_logic():
    run_with_wx(_test_RangeSliderSpinPanel_logic)
def _test_RangeSliderSpinPanel_logic():
    frame = wx.GetApp().GetTopWindow()
    panel = rangeslider.RangeSliderSpinPanel(frame,
                                             lowLabel='Low',
                                             highLabel='High')
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()
    _test_RangePanel_logic(panel)


def test_RangeSliderSpinPanel_onchange():
    run_with_wx(_test_RangeSliderSpinPanel_onchange)
def _test_RangeSliderSpinPanel_onchange():
    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = rangeslider.RangeSliderSpinPanel(frame,
                                             lowLabel='Low',
                                             highLabel='High',
                                             style=0)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()

    called = [False]

    def handler(ev):
        called[0] = True

    panel.Bind(rangeslider.EVT_LOW_RANGE,  handler)
    panel.Bind(rangeslider.EVT_HIGH_RANGE, handler)
    panel.Bind(rangeslider.EVT_RANGE,      handler)

    panel.SetRange(0, 100)
    panel.SetDistance(5)

    realYield()

    panel.lowSpin.textCtrl.ChangeValue('75')
    simkey(sim, panel.lowSpin.textCtrl, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (75, 100)))
    assert np.isclose(panel.lowSpin.GetValue(),   75)
    assert np.isclose(panel.lowSlider.GetValue(), 75)
    assert called[0]

    called[0] = False
    panel.highSpin.textCtrl.ChangeValue('60')
    simkey(sim, panel.highSpin.textCtrl, wx.WXK_RETURN)
    assert np.all(np.isclose(panel.GetRange(), (55, 60)))
    assert np.isclose(panel.highSpin.GetValue(),   60)
    assert np.isclose(panel.highSlider.GetValue(), 60)
    assert called[0]

    called[0] = False
    simclick(sim, panel.lowSlider, pos=[0.25, 0.5])
    assert abs(panel          .GetLow()   - 25) < 5
    assert abs(panel.lowSpin  .GetValue() - 25) < 5
    assert abs(panel.lowSlider.GetValue() - 25) < 5
    assert called[0]

    called[0] = False
    simclick(sim, panel.highSlider, pos=[0.90, 0.5])
    assert abs(panel           .GetHigh()  - 90) < 5
    assert abs(panel.highSpin  .GetValue() - 90) < 5
    assert abs(panel.highSlider.GetValue() - 90) < 5
    assert called[0]

def test_RangeSliderSpinPanel_onlimit():
    run_with_wx(_test_RangeSliderSpinPanel_onlimit)
def _test_RangeSliderSpinPanel_onlimit():

    sim   = wx.UIActionSimulator()
    frame = wx.GetApp().GetTopWindow()
    panel = rangeslider.RangeSliderSpinPanel(
        frame,
        style=rangeslider.RSSP_SHOW_LIMITS | rangeslider.RSSP_EDIT_LIMITS)
    sizer = wx.BoxSizer(wx.HORIZONTAL)
    sizer.Add(panel, flag=wx.EXPAND, proportion=1)
    frame.SetSizer(sizer)
    frame.Layout()

    panel.SetLimits(0, 100)
    panel.SetRange( 0, 100)

    called = [False]

    def handler(ev):
        called[0] = True

    panel.Bind(rangeslider.EVT_RANGE_LIMIT, handler)

    numberdlg = mock.MagicMock()
    numberdlg.ShowModal.return_value = wx.ID_OK
    numberdlg.GetValue .return_value = 9

    minbtn = panel.minButton
    maxbtn = panel.maxButton

    # limbutton, value, expectedrange, shouldTriggerEvent
    testcases = [
        (minbtn,   50, (  50, 100),  True),
        (minbtn,  150, (  50, 100), False),
        (minbtn, -100, (-100, 100),  True),
        (maxbtn,   50, (-100,  50),  True),
        (maxbtn, -200, (-100,  50), False),
        (maxbtn,  500, (-100, 500),  True),
    ]

    with mock.patch('fsleyes_widgets.rangeslider.numberdialog.NumberDialog',
                    return_value=numberdlg):

        for btn, val, expected, shouldEv in testcases:

            called[0] = None

            numberdlg.GetValue.return_value = val

            simclick(sim, btn)
            assert tuple(panel.GetLimits()) == expected

            if shouldEv: assert     called[0]
            else:        assert not called[0]



#!/usr/bin/env python
#
# __init__.py - fsleyes-widgets
#
# Author: Paul McCarthy <pauldmccarthy@gmail.com>
#
"""The ``fsleyes_widgets`` package contains various custom :mod:`wx` widgets
and utilities used by FSLeyes.


Some of the controls in ``fsleyes_widgets`` are duplicates of controls which
are already available in ``wx`` or ``wx.lib.agw``. In these instances, I wrote
my own implementations to work around annoying, quirky, and/or downright buggy
behaviour in the existing controls.
"""


__version__ = '0.0.2'
